/********************************************************
 * Project          : Temperature Datalogger            *
 * Version          : 0.1                               *
 * Date             : 4/18/2000                         *
 * Author           : Mohamed Nazlen Abdul Jabar (NAZ)	*
 * Company          : Cornell University                *
 * Chip type        : ATmega103L                        *
 * Clock frequency  : 4.000000 MHz                      *
 * Memory model     : Small                             *
 * Internal RAM size: 4096                              *
 * External RAM size: 0                                 *
 * Data Stack size  : 1024                              *
*********************************************************/

#include <mega103.h>
#include <stdlib.h>  
#include <string.h>
#include <ftoa.h>
#asm
.equ	__lcd_port=0x18		; PORTB
#endasm
#include <lcd.h>    
#define max_data 2500    

// Subroutine Declaration
char not_leap(void);
void dump_mem(void);
                         
// Structure Declaration
struct ram_structure 
{
	unsigned char second;
	unsigned char minute;
	unsigned char hour;
	unsigned char date;
	unsigned char month;
	 unsigned int year;			
};

struct temp_structure
{
           struct ram_structure tstamp;
	        float  tempread;
	unsigned char end_status;
};                 
                  
// Global Variable
          bit  start_aq=0;
          bit  disp_status=0;
          bit  flop=0;
         char  stemp[10];                        
        float  temperature;
       struct  ram_structure time;
 unsigned int  index_data;
       struct  temp_structure tempsave[max_data];
 unsigned int  int_sec;
 unsigned int  int_min;
 unsigned int  int_hour;
 unsigned int  Tinterval;
unsigned char  Tintsel;    
unsigned char  Tchannel;
         char  busy;
          bit  busyS;
 
sfrw ADCW=4;
#define FIRST_ADC_INPUT 0
#define LAST_ADC_INPUT 1
unsigned int adc_data[LAST_ADC_INPUT-FIRST_ADC_INPUT+1];

#pragma used+
// UART Receiver buffer
char rx_buffer[8];
unsigned char rx_wr_index,rx_rd_index,rx_counter;
// This flag is set on UART Receiver buffer overflow
bit rx_buffer_overflow;
#pragma used-

// ADC interrupt service routine
// with auto input scanning
#pragma savereg-
interrupt [ADC] void adc_isr(void)
{
#asm
    push r26
    push r27
    push r30
    push r31
    in   r30,sreg
    push r30
#endasm
static unsigned char input_index;
// Read the ADC conversion result
adc_data[input_index]=ADCW;
// Select next ADC input
if (++input_index>(LAST_ADC_INPUT-FIRST_ADC_INPUT+1))
   input_index=0;
#ifdef BANDGAP
ADMUX=input_index+FIRST_ADC_INPUT|0x40;
#else
ADMUX=input_index+FIRST_ADC_INPUT;
#endif
// Start the next ADC conversion
ADCSR.6=1;
#asm
    pop  r30
    out  sreg,r30
    pop  r31
    pop  r30
    pop  r27
    pop  r26
#endasm
}
#pragma savereg+

// UART Receiver interrupt service routine
#pragma savereg-
interrupt [UART_RXC] void uart_rx_isr(void)
{
#asm
    .equ __rx_buffer_size=8
    push r26
    in   r26,sreg
    push r26
    push r27
    push r30
#endasm
#ifdef _MODEL_TINY_
#asm
    ldi  r26,_rx_buffer
    lds  r30,_rx_wr_index
    add  r26,r30
    clr  r27
#endasm
#endif
#ifdef _MODEL_SMALL_
#asm
    push r31
    ldi  r26,low(_rx_buffer)
    ldi  r27,high(_rx_buffer)
    lds  r30,_rx_wr_index
    clr  r31
    add  r26,r30
    adc  r27,r31
#endasm
#endif
#asm
    in   r30,udr
    st   x,r30
    lds  r30,_rx_wr_index
    inc  r30
    cpi  r30,__rx_buffer_size
    brlo __uart_rx_isr0
    clr  r30
__uart_rx_isr0:
    sts  _rx_wr_index,r30
    lds  r30,_rx_counter
    inc  r30
    sts  _rx_counter,r30
    cpi  r30,__rx_buffer_size+1
    brlo __uart_rx_isr1
#endasm
rx_buffer_overflow=1;
#asm("__uart_rx_isr1:")
#ifdef _MODEL_SMALL_
#asm("pop  r31")
#endif
#asm
    pop  r30
    pop  r27
    pop  r26
    out  sreg,r26
    pop  r26
#endasm
}
#pragma savereg+

// Get a character from the UART Receiver buffer
#define _ALTERNATE_GETCHAR_
#pragma warn-
#pragma used+
char getchar(void)
{
#asm
    lds  r30,_rx_counter
    tst  r30
    breq _getchar
#endasm
#ifdef _MODEL_TINY_
#asm
    ldi  r26,_rx_buffer
    lds  r30,_rx_rd_index
    add  r26,r30
    clr  r27
#endasm
#endif
#ifdef _MODEL_SMALL_
#asm
    ldi  r26,low(_rx_buffer)
    ldi  r27,high(_rx_buffer)
    lds  r30,_rx_rd_index
    clr  r31
    add  r26,r30
    adc  r27,r31
#endasm
#endif
#asm
    inc  r30
    cpi  r30,__rx_buffer_size
    brlo __getchar0
    clr  r30
__getchar0:
    sts  _rx_rd_index,r30
    ld   r30,x
    cli
    lds  r26,_rx_counter
    dec  r26
    sts  _rx_counter,r26
    sei
#endasm
}
#pragma used-
#ifdef _WARNINGS_ON_
#pragma warn+
#endif


/**********************************************
 * Subroutine : TIMER0 ISR                    *
 * Purpose    : Real Time Clock               *
 *              Check for interval and once   *
 *              due, dump data into memory    *
 **********************************************/
 
interrupt [TIM0_OVF] void timer0_ovf_isr(void)
{   
	// Check for interval(second) and if due, call dump_mem 
	if (start_aq & (Tintsel == 0))
		if (++int_sec == Tinterval) 
		{
			dump_mem(); 	// dump data into memory
		    int_sec = 0;	// Initialize variable
		 }
   	disp_status = !disp_status;   	// Flip display on LCD
   	flop = 1;                       // Set display to true
   	if (++busy > 120)				// No action, clear LCD
   	{                        
   		busy = 0;					// ResSet busy counter 
   		busyS = 0;					// Set Busy Flag to False
   		lcd_clear();				// Clear LCD
   	}
	if (++time.second == 60)
	{   
		time.second = 0;         	// Reset Second
		// Check for interval(second) and if due, call dump_mem
		if (start_aq & (Tintsel == 1))
			if (++int_min == Tinterval) 
			{
				dump_mem();			// dump data into memory
				int_min = 0;		// Reset variable interval for minute
			}	
		if (++time.minute == 60)
		{   
			time.minute = 0;  		// Reset minute
			// Check for interval(second) and if due, call dump_mem
			if (start_aq & (Tintsel == 0))
				if (++int_hour == Tinterval) 
					{
						dump_mem();		// dump data into memory
						int_hour = 0;	// Reset variable interval for hour
					}				
			if (++time.hour == 24)
			{
				time.hour = 0;				// Reset hour
				if (++time.date == 32)
				{
					time.month++;
					time.date = 1;			// Reset date to 1
				}               
				else if (time.date == 31)
				{
					if ((time.month==4) | (time.month==6) | (time.month==9) | (time.month==11))
					{
						time.month++;
						time.date = 1; 		// Reset date to 1
					}
				}
				else if (time.date == 30)
				{
					if (time.month == 2)
					{
						time.month++;   	// Increase month by 1
						time.date = 1; 		// Reset date to 1
					}					
				}
				else if (time.date == 29)
				{
					if ((time.month == 2) && (not_leap()))
					{
						time.month++;    	// Increase month by 1
						time.date = 1;  	// Reset date to 1
					}
				}    
				if (time.month == 13)
				{
					time.month = 1;     	// Reset month tp 1
					time.year++;			// Increase year by 1
				}	// IF Month
			}	// IF Second
		}	// IF Minute
	}	// IF Second
}	// Interrupt Timer0

#include <stdio.h>
#include <delay.h>
	
void main(void)
{   
	// Local Variable delcaration
	         char  lcd_buffer[8];
	         char  lcd_buffer2[8];
	         char  k; 
	         char  TempChar[5];                
	          int  i;                     	
	unsigned char  Tsecond;
	unsigned char  Tminute;
	unsigned char  Thour;
	unsigned char  Tdate;
	unsigned char  Tmonth;
	unsigned int   Tyear;
			 char  Gchannel;
	           
	// Input/Output Ports initialization
	DDRB=0xFF;				// Port B
	PORTB=0xFF;
	DDRD=0x00;				// Port D
	PORTD=0x00;

	// Timer/Counter 0 initialization
	TCCR0=0x05;				// Clock value: PCK0/128
	ASSR=0x08;				// Clock source: TOSC1 pin
	TCNT0=0x00;				// Mode: Output Compare
	OCR0=0x00;				// OC0 output: Disconnected

	// Timer(s)/Counter(s) Interrupt(s) initialization
	TIMSK=0x01;

	// UART initialization
	// Communication Parameters: 8 Data, 1 Stop, No Parity
	// UART Receiver: On
	// UART Transmitter: On
	UCR=0x98;
	UBRR=0x0C;					// UART Baud rate: 19200

	// ADC initialization
	// ADC Clock frequency: 31.250 kHz
	ADMUX=FIRST_ADC_INPUT;
	ADCSR=0xCF;

	// Reset all variable to appropriate value
	  time.hour = 0;
	time.minute = 0;
	time.second = 0;
	  time.date = 1;	// Set initial value for date = 01/01/2000
	 time.month = 1;
	  time.year = 2000;
	 index_data = 0;
   	    int_sec = 0;
	    int_min = 0;
	   int_hour = 0;
     index_data = 0;
	      busyS = 1;	
     
	// LCD initialization 
	lcd_init(16);   	
	// Display "TEMPREATURE LOGG" on LCD
	lcd_clear();         
	lcd_gotoxy(0,0);
	lcd_putsf("TEMPERAT");	
	_lcd_ready(); 
	_lcd_write_data(0xC0);
	lcd_putsf("URE LOGG");   
	delay_ms(1500);                     
	// Display "BY NAZ @ CORNELL" on LCD
	lcd_clear();         
	lcd_gotoxy(0,0);
	lcd_putsf("BY NAZ @");	
	_lcd_ready(); 
	_lcd_write_data(0xC0);
	lcd_putsf(" CORNELL");   
	delay_ms(1000);	                        	
	// Initialize all record
	for (i=0;i<max_data;i++)                             	
	{
		tempsave[i].tstamp.second = 0;
		tempsave[i].tstamp.minute = 0;
		tempsave[i].tstamp.hour   = 0;
		tempsave[i].tstamp.year   = 2000;
		tempsave[i].tstamp.month  = 0;
		tempsave[i].tstamp.date   = 0;
		tempsave[i].tempread      = 0.0;
		tempsave[i].end_status     = 0;
	};	
	// set the first record as end
	tempsave[0].end_status = 1;	   							
	// Global enable interrupts
	
	#asm("sei")
	
	while (1)         
	{                 
	
	if (PINE.4 == 0)			// Check for Push Button
	{
		busyS = 1;				// On busy flag
		busy = 0;				// Reset busy counter		
	}
  	if (rx_counter != 0)		// Check for any character on Receive 
    	{                       // 	    buffer
      		busyS = 1;          // On busy flag
      		busy = 0;			// Reset busy counter
      		k = getchar();
			if (k = '%')        // Check for start character "%"
			{
				k = getchar();	// Get the second character
				switch (k)		// Take necessary action based on 2nd char
				{
					case '0': 	// Syncronize Time
								// Hour
								TempChar[0] = getchar(); 									
								TempChar[1] = getchar(); 										
								TempChar[2] = '\0';
								Thour = atoi(TempChar);
								// Minute
								TempChar[0] = getchar(); 									
								TempChar[1] = getchar(); 										
								TempChar[2] = '\0';
								Tminute = atoi(TempChar);
								// Second
								TempChar[0] = getchar(); 									
								TempChar[1] = getchar(); 										
								TempChar[2] = '\0';
								Tsecond = atoi(TempChar);
								// Year
								TempChar[0] = getchar(); 										
								TempChar[1] = getchar(); 									
								TempChar[2] = getchar(); 										
								TempChar[3] = getchar(); 		
								TempChar[4] = '\0';
								Tyear = atoi(TempChar);
								// Month
								TempChar[0] = getchar(); 										
								TempChar[1] = getchar(); 											
								TempChar[2] = '\0';
								Tmonth = atoi(TempChar);
								// Date
								TempChar[0] = getchar(); 										
								TempChar[1] = getchar(); 										
								TempChar[2] = '\0';
								Tdate = atoi(TempChar);
								// Uploading the setting into global memory																
								#asm("cli")
								time.hour = Thour;
								time.minute = Tminute;
								time.second = Tsecond;
								time.year = Tyear;
								time.month = Tmonth;
								time.date = Tdate;
								#asm("sei")	   
								rx_counter = 0;							
								break;
					case '1':	// Setting
								// Interval
								TempChar[0] = getchar(); 											
								TempChar[1] = getchar(); 										
								TempChar[2] = '\0';
								Tinterval = atoi(TempChar);
								// Interval Select
								TempChar[0] = getchar(); 								
								TempChar[1] = '\0';
								Tintsel = atoi(TempChar);
								// Channel
								TempChar[0] = getchar(); 									
								TempChar[1] = '\0';
								Tchannel = atoi(TempChar);	
								rx_counter = 0;								
								break;
					case '2':	// Upload Data             
								for (i=1; i<max_data;i++)
								{   
									ftoa(tempsave[i].tempread,2,stemp);
									printf("&2%-d:%-d:%-d,%-d/%-d/%-d,%s%%\r\n",
												tempsave[i].tstamp.hour,
												tempsave[i].tstamp.minute,
												tempsave[i].tstamp.second,
												tempsave[i].tstamp.month,
												tempsave[i].tstamp.date,
												tempsave[i].tstamp.year,
										        stemp); 
							 	 	TempChar[0] = getchar();
							 	 	if (TempChar[0] != '#' )
							 	 		i--;   		 	 								 	 	
								 }	 
								break;
					case '3':	// Reset Data
								lcd_clear();
								lcd_gotoxy(0,0);
			   					lcd_putsf("Reseting ...");
			   					// Initialize all record								
								for (i=0;i<max_data;i++)                             		
								{
									tempsave[i].tstamp.second = 0;
									tempsave[i].tstamp.minute = 0;
									tempsave[i].tstamp.hour   = 0;
									tempsave[i].tstamp.year   = 0;
									tempsave[i].tstamp.month  = 0;
									tempsave[i].tstamp.date   = 0;
									tempsave[i].tempread      = 0.0;
									tempsave[i].end_status     = 0;
								};
								// set the first record as end
								tempsave[0].end_status = 1;	   								
								index_data = 0;   
								break; 
					case '4':   // Start Aquiring
								start_aq = 1;                 
								int_sec = 0;
								int_min = 0;
								int_hour = 0;  
					            break;
					case '5':   // Stop Aquiring
								start_aq = 0;  
					            break;                       
					case '6':	// Download Current Temperature  
								lcd_clear();
								lcd_gotoxy(0,0);
								lcd_putsf("Upload..");
								TempChar[0] = getchar();
								TempChar[1] = '\0';
								Gchannel = atoi(TempChar);
								temperature = (adc_data[Gchannel] * 300.0)/1024.0;
						 		ftoa(temperature,2,stemp);
						 		if (Gchannel == 0)
							   		printf("%s'F\r\n",stemp);
							  	else
							  		printf("%s'C\r\n",stemp);	
						   		break;    
					case '7':	// Test communication
								printf("Test Communication Response\r\n");
								break;
					default :	break;
				}	// switch k
 				rx_counter = 0;		  // Clear the received buffer		
			}   // if k
			rx_counter = 0;			// Clear the received buffer
    	}	// if  rx_counter
    	else
    	{   
    		if (flop & busyS)
    		{
 				if (disp_status)
				{  
					sprintf(lcd_buffer,"%-d:%-d:%-d", time.hour, time.minute, time.second); 
					switch(time.month)
					{
						case 1 : sprintf(lcd_buffer2," Jan %-d",time.date); break; 
						case 2 : sprintf(lcd_buffer2," Feb %-d",time.date); break; 
						case 3 : sprintf(lcd_buffer2," Mar %-d",time.date); break;
						case 4 : sprintf(lcd_buffer2," Apr %-d",time.date); break;
						case 5 : sprintf(lcd_buffer2," May %-d",time.date); break;
						case 6 : sprintf(lcd_buffer2," Jun %-d",time.date); break;
						case 7 : sprintf(lcd_buffer2," Jul %-d",time.date); break;
						case 8 : sprintf(lcd_buffer2," Aug %-d",time.date); break;
						case 9 : sprintf(lcd_buffer2," Sep %-d",time.date); break;
						case 10 : sprintf(lcd_buffer2," Oct %-d",time.date); break;
						case 11 : sprintf(lcd_buffer2," Nov %-d",time.date); break;
						case 12 : sprintf(lcd_buffer2," Dec %-d",time.date); break;
					}
				}
		   		else
				{  
					// Farenheit calculation 
					temperature = (adc_data[0] * 300.0)/1024.0;
			 		ftoa(temperature,2,stemp);      	// Convert value into string
			   		sprintf(lcd_buffer,"%s'F",stemp);  	// put value into LCD_BUFFER
			   		// Celcius calculation 
					temperature = (adc_data[1] * 300.0)/1024.0;
			 		ftoa(temperature,2,stemp);			// Convert value into string
			   		sprintf(lcd_buffer2,"%s'C",stemp);	// put value into LCD_BUFFER2		   		
				};
				lcd_clear();		// clear the LCD
				lcd_gotoxy(0,0);       	// Goto the character position 0
				lcd_puts(lcd_buffer); 	// display the  Farenheit value                                                
		   		_lcd_ready();           // Wait until LCD ready
		   		_lcd_write_data(0xC1);	// Goto character position 9
		   		lcd_puts(lcd_buffer2);	// Display the Calcius value
		   		flop = 0;				// Reset the flag - Display once every second
		  	} // If Flop
    	}	// else
	}	// While
}   // Main

/****************************************
 * Subroutine : not_leap(void)   		*
 * Purpose    : To check for leap year 	*
 * Return     : True/False				*
 ****************************************/
char not_leap(void)
{
	if (!(time.year%100))
		return(char)(time.year%400);
	else
		return(char)(time.year%4);
}                   


/****************************************
 * Subroutine : dump_mem(void)   		*
 * Purpose    : Save data into SRAM  	*
 * Return     : Void					*
 ****************************************/
void dump_mem(void)
{                          
	// Check if reaching the end.
	// If so ... return to the beginning of record                    
	if (index_data	> max_data-1) 
		{
			index_data = 0;
		    tempsave[max_data-1].end_status = 0;
		 }
	else
		tempsave[index_data-1].end_status = 0;    // Shift end_status into next
	
	// Store appopriate data into memory (SRAM)
	tempsave[index_data].tstamp.second = time.second;
	tempsave[index_data].tstamp.minute = time.minute;
	tempsave[index_data].tstamp.hour = time.hour;
	tempsave[index_data].tstamp.date = time.date;
	tempsave[index_data].tstamp.month = time.month;
	tempsave[index_data].tstamp.year = time.year;
	tempsave[index_data].tempread = (adc_data[Tchannel] * 300.0)/1024.0;
	tempsave[index_data].end_status = 1;         //   Set flag of end status
	index_data++;								// goto next record
}